import requests
import socket


def get_current_ip_address() -> str:
    host = "ifconfig.co"
    addrs = socket.getaddrinfo(host, 80, family=socket.AF_INET, proto=socket.IPPROTO_TCP)
    for addr in addrs:
        ipv4_addr = addr[4][0]
        resp = requests.get("http://" + ipv4_addr + "/ip", headers={"Host": host})
        if resp.ok:
            return resp.text.strip()
    raise Exception("Failed to get current ip address")



import os
import click
import toml
import json
from tabulate import tabulate

from tidbcloudy.api.tidbcloud import TiDBCloud
from tidbcloudy.api.project import Project
from tidbcloudy.api.cluster import Cluster
from tidbcloudy.exception import TiDBCloudResponseException
from tidbcloudy.util.timestamp import timestamp_to_string
from tidbcloudy.api.specification import ClusterType, CreateClusterConfig


@click.group()
@click.option("--profile", default="default", help="The profile name in config file")
@click.pass_context
def cli(ctx, profile):
    ctx.ensure_object(dict)

    obj = ctx.obj
    obj["config_path"] = os.path.expanduser("~/.tidbcloudy.toml")
    obj["profile_name"] = profile
    obj["profile"] = {}
    obj["api"] = None
    obj["project"] = None

    if os.path.isfile(obj["config_path"]):
        config_content = toml.load(obj["config_path"])
        if profile in config_content:
            obj["profile"] = config_content[profile]
            obj["api"] = TiDBCloud(
                public_key=obj["profile"].get("public_key"),
                private_key=obj["profile"].get("private_key"),
                base_url=obj["profile"].get("base_url"))
            obj["project"] = obj["api"].get_project(obj["profile"].get("project_id"))


@cli.group(name="cluster")
def cli_cluster():
    pass


@cli_cluster.command("create")
@click.option("-f", "--file", help="The json configuration file")
@click.pass_obj
def cli_cluster_create(obj, f):
    cluster_config = CreateClusterConfig()
    api = obj["api"]  # type: TiDBCloud
    project = obj["project"]  # type: Project
    if f is not None:
        cluster_config = json.load(f)
        new_cluster = project.create_cluster(cluster_config)
        click.echo(f"Creating the cluster...")
        new_cluster.wait_for_available()
        return
    cluster_type = click.prompt("Select cluster type",
                                type=click.Choice([str(item.value) for item in ClusterType], case_sensitive=False))
    cluster_config.set_cluster_type(cluster_type)
    name = click.prompt("Specify the cluster name", type=str)
    cluster_config.set_name(name)
    if cluster_type == "DEDICATED":
        port = click.prompt("Specify the cluster port", type=int)
        cluster_config.set_port(port)
    click.echo("Loading cloud provider...")
    provider_regions = api.list_provider_regions()
    available_provider = set()
    for item in provider_regions:
        if item.cluster_type.value == cluster_type:
            available_provider.add(str(item.cloud_provider.value))
    cloud_provider = click.prompt("Select cloud provider",
                                  type=click.Choice(sorted(available_provider), case_sensitive=False))
    cluster_config.set_cloud_provider(cloud_provider)
    click.echo("Loading regions...")
    available_regions = set()
    for item in provider_regions:
        if item.cluster_type.value == cluster_type and item.cloud_provider.value == cloud_provider:
            available_regions.add(str(item.region))
    region = click.prompt("Select cloud region", type=click.Choice(sorted(available_regions), case_sensitive=False))
    cluster_config.set_region(region)
    if cluster_type == "DEDICATED":
        click.echo("Loading cluster configurations...")
        for component in ("tidb", "tikv", "tiflash"):
            for item in provider_regions:
                if item.cluster_type.value == cluster_type and item.cloud_provider.value == cloud_provider and \
                        item.region == region:
                    component_config = getattr(item, component)
                    click.echo(f"Specify the size of {component}")
                    # Configure node_size
                    available_node_size = set()
                    for config_item in component_config:
                        available_node_size.add(str(config_item.node_size))
                    node_size = click.prompt("Select the node size",
                                             type=click.Choice(sorted(available_node_size), case_sensitive=False))

                    # Configure node_quantity
                    node_quantity = None
                    node_storage = None
                    for config_item in component_config:
                        if config_item.node_size == node_size:
                            quantity_min = config_item.node_quantity_range.min
                            quantity_step = config_item.node_quantity_range.step
                            node_quantity = click.prompt(
                                f"Specify the node quantity [min={quantity_min} step={quantity_step}]",
                                type=click.IntRange(quantity_min))

                            # Configure node_storage
                            if hasattr(config_item, "storage_size_gib_range"):
                                storage_min = config_item.storage_size_gib_range.min
                                storage_max = config_item.storage_size_gib_range.max
                                node_storage = click.prompt(
                                    f"Specify the node storage [min={storage_min} max={storage_max}]",
                                    type=click.IntRange(storage_min, storage_max))
                    cluster_config.set_component(component, node_size, node_quantity, node_storage)
    cluster_config.set_root_password("THIS_PASSWORD_IS_NOT_USED")
    new_cluster = project.create_cluster(cluster_config)
    click.echo(f"Creating the cluster {name}")
    new_cluster.wait_for_available()


@cli_cluster.command("modify")
def cli_cluster_modify():
    click.echo('modify cluster')


@cli_cluster.command("list")
@click.pass_obj
def cli_cluster_list(obj):
    project = obj["project"]  # type: Project
    cluster_table = []
    for cluster in project.iter_clusters():
        cluster_table.append(
            [cluster.id,
             cluster.name,
             cluster.status.cluster_status.value,
             cluster.cluster_type.value,
             cluster.status.tidb_version,
             f"{cluster.cloud_provider.value}:{cluster.region}",
             timestamp_to_string(cluster.create_timestamp)
             ]
        )
    click.echo(
        tabulate(cluster_table, headers=["ID", "NAME", "STATUS", "TYPE", "VERSION", "PROVIDER", "CREATED"]))


@cli_cluster.command("info")
@click.argument("cluster_id")
@click.pass_obj
def cli_cluster_info(obj, cluster_id):
    project = obj["project"]
    cluster_info = project.get_cluster(cluster_id).to_object()
    click.echo(json.dumps(cluster_info, indent=2))


@cli_cluster.command("delete")
@click.argument("cluster_id")
@click.pass_obj
def cli_cluster_delete(obj, cluster_id):
    project = obj["project"]  # type: Project
    # Confirm
    project.delete_cluster(cluster_id)


@cli.group(name="project")
def cli_project():
    pass


@cli_project.command("list", help="List all projects")
@click.pass_obj
def cli_project_list(obj):
    api = obj["api"]  # type: TiDBCloud
    project_table = []
    for project in api.iter_projects():
        project_table.append(
            [project.id,
             project.name,
             project.org_id,
             project.cluster_count,
             project.user_count,
             timestamp_to_string(project.create_timestamp)
             ]
        )
    click.echo(
        tabulate(project_table, headers=["ID", "NAME", "ORGANIZATION ID", "CLUSTER", "USER", "CREATED"]))


@cli_project.command("info", help="Get the info of a project")
@click.argument("project_id")
@click.pass_obj
def cli_project_info(obj, project_id):
    api = obj["api"]  # type: TiDBCloud
    cluster_info = api.get_project(project_id, update_from_server=True).to_object()
    click.echo(json.dumps(cluster_info, indent=2))


@cli.group(name="backup")
def cli_backup():
    pass


@cli_backup.command("create")
@click.pass_obj
def cli_backup_create():
    click.echo('create backup')


@cli_backup.command("list")
@click.argument("cluster_id")
@click.pass_obj
def cli_backup_list(obj, cluster_id):
    project = obj["project"]  # type: Project
    cluster = project.get_cluster(cluster_id)  # type: Cluster
    backup_table = []
    for backup in cluster.iter_backups():
        backup_table.append(
            [backup.id,
             backup.name,
             backup.description,
             backup.type,
             backup.size,
             backup.create_timestamp
             ]
        )
    click.echo(
        tabulate(backup_table, headers=["ID", "NAME", "DESCRIPTION", "TYPE", "SIZE", "CREATED"]))


@cli_backup.command("info")
@click.argument("cluster_id")
@click.argument("backup_id")
@click.pass_obj
def cli_backup_info(obj, cluster_id, backup_id):
    project = obj["project"]  # type: Project
    cluster = project.get_cluster(cluster_id)  # type: Cluster
    backup_info = cluster.get_backup(backup_id).to_object()
    click.echo(json.dumps(backup_info, indent=2))


@cli_backup.command("delete")
def cli_backup_delete():
    click.echo('delete backup')


@cli.group(name="restore")
def cli_restore():
    pass


@cli_restore.command("create")
def cli_restore_create():
    click.echo('create restore')


@cli_restore.command("list")
@click.pass_obj
def cli_restore_list(obj):
    click.echo('list restore')
    project = obj["project"]
    for restore in project.iter_restores():
        print(restore)


@cli_restore.command("info")
def cli_restore_info():
    click.echo('info restore')


@cli.command("config", help="Configure your TiDB Cloud API key and in ~/.tidbcloudy.toml")
@click.pass_obj
def cli_config(obj):
    profile_name = obj["profile_name"]
    profile = obj["profile"]

    # Read from user input
    public_key = click.prompt('TiDB Cloud API public key',
                              hide_input=False, default=profile.get("public_key"))
    private_key = click.prompt('TiDB Cloud API private key',
                               hide_input=False, default=profile.get("private_key"))
    base_url = click.prompt('TiDB Cloud API base URL [https://api.tidbcloud.com/api/v1beta/]',
                            hide_input=False, default=profile.get("base_url", ""), show_default=False)

    # Load project info from server
    click.echo("Loading projects ...")
    if base_url == "":
        base_url = None
    api = TiDBCloud(public_key=public_key, private_key=private_key, base_url=base_url)
    try:
        projects = list(api.iter_projects())
    except TiDBCloudResponseException as e:
        click.echo("Error: {}".format(e))
        return

    # Select project
    selected_project = ""
    if len(projects) == 0:
        click.echo("No project found.")
    elif len(projects) == 1:
        selected_project = projects[0]
    else:
        click.echo("Available projects:")
        for index, project in enumerate(projects):
            click.echo(f"{index + 1}. {project.name} (id={project.id})")
        project_choice = click.prompt(f'Select a project (1-{len(projects)})', hide_input=False,
                                      type=click.IntRange(1, len(projects)))
        selected_project = projects[project_choice - 1]
    click.echo(f"Selected project: {selected_project.name} (id={selected_project.id})")

    # Create new profile content
    new_profile = {
        "public_key": public_key,
        "private_key": private_key
    }
    if base_url:
        new_profile["base_url"] = base_url
    if selected_project:
        new_profile["project_id"] = selected_project.id

    # Update configuration file
    config_path = obj["config_path"]
    config_content = {}
    if os.path.isfile(config_path):
        config_content = toml.load(config_path)
    config_content[profile_name] = new_profile
    with open(config_path, 'w') as f:
        toml.dump(config_content, f)


if __name__ == "__main__":
    cli()



class TiDBCloudException(Exception):
    pass


class TiDBCloudResponseException(TiDBCloudException):
    def __init__(self, status, code=None, message=None, raw=None):
        self._status = status
        self._code = code
        self._message = message
        self._raw = raw

    @property
    def status(self):
        return self._status

    @property
    def code(self):
        return self._code

    @property
    def message(self):
        return self._message

    @property
    def raw(self):
        return self._raw

    def __str__(self):
        return "status: {}, code: {}, message: {}".format(self._status, self._code, self._message)



import time
import deprecation
import MySQLdb
from typing import Union, Iterator

from tidbcloudy.api._base import TiDBCloudyBase, TiDBCloudyContextualBase, TiDBCloudyField
from .specification import ClusterType, CloudProvider, ClusterConfig, ClusterInfo, UpdateClusterConfig, \
    ClusterStatus
from .backup import Backup
from tidbcloudy.util.log import log
from tidbcloudy.util.timestamp import timestamp_to_string
from tidbcloudy.util.page import Page


# noinspection PyShadowingBuiltins
class Cluster(TiDBCloudyBase, TiDBCloudyContextualBase):
    __slots__ = ["_id", "_project_id", "_name", "_cluster_type", "_cloud_provider", "_region", "_create_timestamp",
                 "_config", "_status"]

    id: str = TiDBCloudyField(str)
    project_id: str = TiDBCloudyField(str)
    name: str = TiDBCloudyField(str)
    cluster_type: ClusterType = TiDBCloudyField(ClusterType)
    cloud_provider: CloudProvider = TiDBCloudyField(CloudProvider)
    region: str = TiDBCloudyField(str)
    create_timestamp: int = TiDBCloudyField(int, convert_from=int, convert_to=str)
    config: ClusterConfig = TiDBCloudyField(ClusterConfig)
    status: ClusterInfo = TiDBCloudyField(ClusterInfo)

    def _update_info_from_server(self):
        path = "projects/{}/clusters/{}".format(self.project_id, self.id)
        resp = self.context.call_get(path=path)
        self.assign_object(resp)

    def wait_for_available(self, *, timeout_sec: int = None, interval_sec: int = 1) -> bool:
        """
        Wait for cluster to be ready.
        Args:
            timeout_sec: timeout in seconds.
            interval_sec: interval in seconds.

        Returns:
            True if cluster is ready, False if timeout.

        Examples:
            ```python
            import tidbcloudy
            from tidbcloudy.specification import CreateClusterConfig
            api = tidbcloudy.TiDBCloud(public_key="your_public_key", private_key="your_private_key")
            project = api.get_project(project_id)
            cluster_config = CreateClusterConfig()
            cluster_config.set_name("your_cluster_name").set_cluster_type("your_cluster_type").set_cloud_provider("your_cloud_provider").set_region("your_region").set_port(4000).set_root_password("your_root_password")
            cluster = project.create_cluster(cluster_config)
            cluster.wait_for_ready()
        """
        time_start = time.monotonic()
        counter = 1
        while True:
            duration = time.monotonic() - time_start
            minutes = duration - 60 * counter
            if timeout_sec is not None and duration > timeout_sec:
                return False
            elif minutes > 0:
                counter += 1
                log("Waiting for cluster {} to be ready, {} seconds passed...".format(self.id, int(duration)))
            self._update_info_from_server()
            if self.status.cluster_status == ClusterStatus.AVAILABLE:
                log("Cluster id={} is {}".format(self.id, self.status.cluster_status.value))
                return True
            time.sleep(interval_sec)

    @deprecation.deprecated(details="Use wait_for_available instead")
    def wait_for_ready(self, *, timeout_sec: int = None, interval_sec: int = 10) -> bool:
        return self.wait_for_available(timeout_sec=timeout_sec, interval_sec=interval_sec)

    def update(self, config: Union[UpdateClusterConfig, dict], update_from_server: bool = False):
        path = "projects/{}/clusters/{}".format(self.project_id, self.id)
        if isinstance(config, UpdateClusterConfig):
            config = config.to_object()
        self.context.call_patch(path=path, json=config)
        log("Cluster id={} has been updated".format(self.id))
        if update_from_server:
            self._update_info_from_server()

    def pause(self):
        path = "projects/{}/clusters/{}".format(self.project_id, self.id)
        config = {
            "config": {
                "paused": True
            }
        }
        self.context.call_patch(path=path, json=config)
        self._update_info_from_server()
        log("Cluster id={} status={}".format(self.id, self.status.cluster_status.value))

    def resume(self):
        path = "projects/{}/clusters/{}".format(self.project_id, self.id)
        config = {
            "config": {
                "paused": False
            }
        }
        self.context.call_patch(path=path, json=config)
        self._update_info_from_server()
        log("Cluster id={} status={}".format(self.id, self.status.cluster_status.value))

    def delete(self):
        path = "projects/{}/clusters/{}".format(self.project_id, self.id)
        self.context.call_delete(path=path)
        log("Cluster id={} has been deleted".format(self.id))

    def create_backup(self, *, name: str, description: str = None) -> Backup:
        """
        Create a backup of the cluster.
        Args:
            name: the name of the backup task.
            description: the description of the backup task.

        Returns:
            Backup instance.

        """
        path = "projects/{}/clusters/{}/backups".format(self.project_id, self.id)
        config = {
            "name": name
        }
        if description is not None:
            config["description"] = description
        resp = self.context.call_post(path=path, json=config)
        return self.get_backup(resp["id"])

    def delete_backup(self, backup_id: str):
        """
        Delete a backup of the cluster.
        Args:
            backup_id: the id of the backup task you want to delete.

        Returns:
            The response of the API.

        """
        Backup(context=self.context, backup_id=backup_id, cluster_id=self.id, project_id=self.project_id).delete()

    def iter_backups(self, *, page_size: int = 10) -> Iterator[Backup]:
        """
        This is not a TiDB Cloud official endpoint.
        Iterate all backups of the cluster.
        Args:
            page_size: the page size of the response.

        Returns:
            Backup instance.

        """
        page = 1
        total = None
        while total is None or (page - 1) * page_size < total:
            backups = self.list_backups(page=page, page_size=page_size)
            total = backups.total
            for backup in backups.items:
                yield backup
            page += 1

    def list_backups(self, *, page: int = None, page_size: int = None) -> Page[Backup]:
        """
        List all backups of the cluster.
        Args:
            page: the page of the response.
            page_size: the page size of each page.

        Returns:
            The response of the API.

        """
        path = "projects/{}/clusters/{}/backups".format(self.project_id, self.id)
        query = {}
        if page is not None:
            query["page"] = page
        if page_size is not None:
            query["page_size"] = page_size
        resp = self.context.call_get(path=path, params=query)
        return Page(
            [Backup.from_object(
                self.context, {"cluster_id": self.id, "project_id": self.project_id, **backup}
            ) for backup in resp["items"]],
            page, page_size, resp["total"]
        )

    def get_backup(self, backup_id: str) -> Backup:
        """
        Get a backup of the cluster.
        Args:
            backup_id: the id of the backup task you want to get.

        Returns:
            Backup instance.

        """
        path = "projects/{}/clusters/{}/backups/{}".format(self.project_id, self.id, backup_id)
        resp = self.context.call_get(path=path)
        return Backup.from_object(self.context, {"cluster_id": self.id, "project_id": self.project_id, **resp})

    def connect(self, type: str, database: str, password: str):
        connection_strings = self.status.connection_strings.to_object()
        user = connection_strings["default_user"]
        if connection_strings is None:
            raise ValueError("No connection strings found for type {}".format(type))
        connection_string = connection_strings[type]
        return MySQLdb.connect(host=connection_string["host"],
                               port=connection_string["port"],
                               user=user, password=password,
                               database=database)

    def __repr__(self):
        if self.status.cluster_status == ClusterStatus.CREATING.value:
            return "<Cluster id={} CREATING>".format(self.id)
        elif self.cluster_type is None:
            return "<Cluster id={} name={}>".format(self.id, self.name)
        else:
            return "<Cluster id={} name={} type={} create_at={}>".format(
                self.id, self.name, self.cluster_type.value,
                timestamp_to_string(self.create_timestamp))



from typing import Union, Iterator

from tidbcloudy.api._base import TiDBCloudyBase, TiDBCloudyContextualBase, TiDBCloudyField
from .cluster import Cluster
from .restore import Restore
from .specification import CreateClusterConfig, UpdateClusterConfig
from tidbcloudy.util.timestamp import timestamp_to_string
from tidbcloudy.util.page import Page


# noinspection PyShadowingBuiltins
class Project(TiDBCloudyBase, TiDBCloudyContextualBase):
    __slots__ = ["_id", "_org_id", "_name", "_cluster_count", "_user_count", "_create_timestamp"]
    id: str = TiDBCloudyField(str)
    org_id: str = TiDBCloudyField(str)
    name: str = TiDBCloudyField(str)
    cluster_count: int = TiDBCloudyField(int)
    user_count: int = TiDBCloudyField(int)
    create_timestamp: int = TiDBCloudyField(int, convert_from=int, convert_to=str)

    def create_cluster(self, config: Union[CreateClusterConfig, dict]) -> Cluster:
        """
        Create a cluster in the project.
        Args:
            config: the configuration of the cluster.

        Returns:
            The created cluster instance.

        Examples:
            .. code-block:: python
                import tidbcloudy
                api = tidbcloudy.TiDBCloud(public_key="your_public_key", private_key="your_private_key")
                project = api.get_project(project_id)
                config = CreateClusterConfig()
                config.set_name("your_cluster_name") \
                        .set_cluster_type("your_cluster_type") \
                        .set_cloud_provider("your_cloud_provider") \
                        .set_region("your_region") \
                        .set_port(4000) \
                        .set_root_password("your_root_password") \
                        .set_component("tidb", "8C16G", 2) \
                        .set_component("tikv", "8C32G", 3, 500) \
                        .set_component("tiflash", "16C128G", 1, 500) \
                        .add_ip_access(cidr="your ip") \
                        .add_ip_access(cidr="your ip")
                cluster = project.create_cluster(config)
                cluster.wait_for_ready()
        """
        if isinstance(config, CreateClusterConfig):
            config = config.to_object()
        path = "projects/{}/clusters".format(self.id)
        resp = self.context.call_post(path=path, json=config)
        return Cluster(context=self.context, id=resp["id"], project_id=self.id)

    def update_cluster(self, cluster_id: str, config: Union[UpdateClusterConfig, dict]):
        """
        Update the cluster.
        Args:
            cluster_id: the id of the cluster you want to update.
            config: the updated configuration of the cluster.

        Returns:
            The updated cluster instance.

        Examples:
            .. code-block:: python
                import tidbcloudy
                api = tidbcloudy.TiDBCloud(public_key="your_public_key", private_key="your_private_key")
                project = api.get_project(project_id)
                cluster = project.get_cluster(cluster_id)
                new_cluster_config = UpdateClusterConfig()
                new_cluster_config.update_component("tidb", 1).update_component("tikv", 1)
                new_cluster = project.update_cluster(cluster_id=cluster._id, config=new_config.to_object())
        """
        Cluster(context=self.context, id=cluster_id, project_id=self.id).update(config)

    def delete_cluster(self, cluster_id: str):
        """
        Delete the cluster.
        Args:
            cluster_id: the id of the cluster you want to delete.

        Returns:
            The response of the delete request.

        Examples:
            .. code-block:: python
                import tidbcloudy
                api = tidbcloudy.TiDBCloud(public_key="your_public_key", private_key="your_private_key")
                project = api.get_project(project_id)
                project.delete_cluster(cluster_id)
        """
        Cluster(context=self.context, id=cluster_id, project_id=self.id).delete()

    def get_cluster(self, cluster_id: str) -> Cluster:
        """
        Get the cluster.
        Args:
            cluster_id: the id of the cluster you want to get.

        Returns:
            The cluster instance.

        Examples:
            .. code-block:: python
                import tidbcloudy
                api = tidbcloudy.TiDBCloud(public_key="your_public_key", private_key="your_private_key")
                project = api.get_project(project_id)
                cluster = project.get_cluster(cluster_id)

        """
        path = "projects/{}/clusters/{}".format(self.id, cluster_id)
        resp = self.context.call_get(path=path)
        return Cluster.from_object(self.context, resp)

    def iter_clusters(self, page_size: int = 10) -> Iterator[Cluster]:
        """
        This is not a TiDB Cloud API official endpoint.
        Iterate all clusters in the project.
        Args:
            page_size:

        Returns:
            The iterator of the clusters.

        Examples:
            .. code-block:: python
                import tidbcloudy
                api = tidbcloudy.TiDBCloud(public_key="your_public_key", private_key="your_private_key")
                project = api.get_project(project_id)
                for cluster in project.iter_clusters():
                    print(cluster) # This is a Cluster instance.

        """
        page = 1
        total = None
        while total is None or (page - 1) * page_size < total:
            clusters = self.list_clusters(page=page, page_size=page_size)
            total = clusters.total
            for cluster in clusters.items:
                yield cluster
            page += 1

    def list_clusters(self, page: int = None, page_size: int = None) -> Page[Cluster]:
        """
        List all clusters in the project.
        Args:
            page:
            page_size:

        Returns:

        Examples:
            .. code-block:: python
                import tidbcloudy
                api = tidbcloudy.TiDBCloud(public_key="your_public_key", private_key="your_private_key")
                project = api.get_project(project_id)
                clusters = project.list_clusters()
                for cluster in clusters["items"]:
                    print(cluster) # This is a Cluster instance.

        """
        path = "projects/{}/clusters".format(self.id)
        query = {}
        if page is not None:
            query["page"] = page
        if page_size is not None:
            query["page_size"] = page_size
        resp = self.context.call_get(path=path, params=query)
        return Page(
            [Cluster.from_object(self.context, item) for item in resp["items"]],
            page, page_size, resp["total"])

    def create_restore(self, *, name: str, backup_id: str, cluster_config: Union[CreateClusterConfig, dict]) -> Restore:
        """
        Create a restore in the project.
        Args:
            name:
            backup_id:
            cluster_config:

        Returns:

        """
        path = "projects/{}/restores".format(self.id)
        if isinstance(cluster_config, CreateClusterConfig):
            cluster_config = cluster_config.to_object()
        create_config = {
            "name": name,
            "backup_id": backup_id,
            "config": cluster_config["config"]
        }
        resp = self.context.call_post(path=path, json=create_config)
        return Restore(context=self.context, id=resp["id"], cluster_id=resp["cluster_id"])

    def get_restore(self, restore_id: str) -> Restore:
        """
        Get the restore.
        Args:
            restore_id:

        Returns:

        """
        path = "projects/{}/restores/{}".format(self.id, restore_id)
        resp = self.context.call_get(path=path)
        return Restore.from_object(self.context, resp)

    def list_restores(self, *, page: int = None, page_size: int = None) -> Page[Restore]:
        """
        List all restores in the project.
        Args:
            page:
            page_size:

        Returns:

        """
        path = "projects/{}/restores".format(self.id)
        query = {}
        if page is not None:
            query["page"] = page
        if page_size is not None:
            query["page_size"] = page_size
        resp = self.context.call_get(path=path, params=query)
        return Page(
            [Restore.from_object(self.context, item) for item in resp["items"]],
            page, page_size, resp["total"]
        )

    def iter_restores(self, page_size: int = 10) -> Iterator[Restore]:
        """
        This is not a TiDB Cloud API official endpoint.
        Iterate all restores in the project.
        Args:
            page_size:

        Returns:

        """
        page = 1
        total = None
        while total is None or (page - 1) * page_size < total:
            restores = self.list_restores(page=page, page_size=page_size)
            total = restores.total
            for restore in restores.items:
                yield restore
            page += 1

    def __repr__(self):
        return "<Project id={} name={} create_at={}>".format(
            self.id, self.name, timestamp_to_string(self.create_timestamp))



from requests.auth import HTTPDigestAuth
import requests
from ..exception import TiDBCloudResponseException


class Context:
    def __init__(self,
                 public_key: str,
                 private_key: str,
                 *,
                 base_url: str
                 ):
        """
        Args:
            public_key: your public key to access to TiDB Cloud
            private_key: your private key to access to TiDB Cloud
            base_url: the base_url of TiDB Cloud API, you can change this for internal testing.
        """
        self._session = requests.Session()
        self._session.auth = HTTPDigestAuth(public_key, private_key)
        self._base_url = base_url
        if self._base_url[-1] != "/":
            self._base_url += "/"

    def _call_api(self, method: str, path: str, **kwargs) -> dict:
        resp = self._session.request(method=method, url=self._base_url + path, **kwargs)
        if resp.ok:
            return resp.json()
        try:
            content = resp.json()
            raise TiDBCloudResponseException(
                resp.status_code, content.get("code"),
                resp.reason if content.get("message") is None else content["message"])
        except requests.exceptions.JSONDecodeError:
            content = resp.text
            raise TiDBCloudResponseException(resp.status_code, message=content)

    def call_get(self, path: str, *, params: dict = None) -> dict:
        resp = self._call_api(method="GET", path=path, params=params)
        return resp

    def call_post(self, path: str, *, data: dict = None, json: dict = None) -> dict:
        resp = self._call_api(method="POST", path=path, data=data, json=json)
        return resp

    def call_patch(self, path: str, *, data: dict = None, json: dict = None) -> dict:
        resp = self._call_api(method="PATCH", path=path, data=data, json=json)
        return resp

    def call_delete(self, path) -> dict:
        resp = self._call_api(method="DELETE", path=path)
        return resp



import datetime


def timestamp_to_string(timestamp: int) -> str:
    """
    Convert timestamp to datetime string.
    Args:
        timestamp:

    Returns:

    """
    if timestamp is None:
        return ""
    return datetime.datetime.fromtimestamp(timestamp).isoformat()



from tidbcloudy.api._base import TiDBCloudyBase, TiDBCloudyContextualBase, TiDBCloudyField
from tidbcloudy.api.specification import RestoreStatus, ClusterInfoOfRestore


# noinspection PyShadowingBuiltins
class Restore(TiDBCloudyBase, TiDBCloudyContextualBase):
    __slots__ = ["_id", "_create_timestamp", "_backup_id", "_cluster_id", "_cluster", "_status", "_error_message"]
    id: str = TiDBCloudyField(str)
    create_timestamp: str = TiDBCloudyField(str)
    backup_id: str = TiDBCloudyField(str)
    cluster_id: str = TiDBCloudyField(str)
    cluster: ClusterInfoOfRestore = TiDBCloudyField(ClusterInfoOfRestore)
    status: RestoreStatus = TiDBCloudyField(RestoreStatus)
    error_message: str = TiDBCloudyField(str)

    def __repr__(self):
        if self.status is None:
            return "<restore id={} backup_id={} create_at= {}>".format(self.id, self.backup_id,
                                                                       self.create_timestamp)
        else:
            return "<restore id={} backup_id={} create_at= {} status={}>".format(self.id, self.backup_id,
                                                                                 self.create_timestamp,
                                                                                 self.status.value)



import sys

from .api.tidbcloud import TiDBCloud
from .api.context import Context
from .api.project import Project
from .api.cluster import Cluster
from .api.backup import Backup
from .api.restore import Restore
from .exception import TiDBCloudException

from .api import backup
from .api import cluster
from .api import context
from .api import project
from .api import restore
from .api import specification
from .api import tidbcloud

_aliases = ["backup", "cluster", "context", "project", "restore", "specification", "tidbcloud"]

for _alias in _aliases:
    sys.modules["tidbcloudy.{}".format(_alias)] = sys.modules["tidbcloudy.api.{}".format(_alias)]



from typing import Iterator, List, Optional

from tidbcloudy.api.context import Context
from tidbcloudy.api.project import Project
from tidbcloudy.api.specification import CloudSpecification
from tidbcloudy.util.page import Page


class TiDBCloud:

    DEFAULT_BASE_URL = "https://api.tidbcloud.com/api/v1beta/"

    def __init__(self, public_key: str, private_key: str, *, base_url: Optional[str] = None):
        if base_url is None:
            base_url = self.DEFAULT_BASE_URL
        self._context = Context(public_key, private_key, base_url=base_url)

    def get_project(self, project_id: str, update_from_server: bool = False) -> Project:
        """
        Get the project object by project_id.
        Args:
            project_id: the project id.
            update_from_server: whether to update the project info.

        Returns:
            If the update_from_server is False, return a Project object with only the context and project_id.
            If the update_from_server is True, return a Project object with all the info.

        Examples:
            .. code-block:: python
                import tidbcloudy
                api = tidbcloudy.TiDBCloud(public_key="your_public_key", private_key="your_private_key")
                project = api.get_project("your_project_id", update_from_server=True)
        """
        project = Project(context=self._context, id=project_id)
        if update_from_server:
            for item in self.iter_projects():
                if item.id == project_id:
                    return item
            raise ValueError("Project {} not found".format(project_id))
        return project

    def list_projects(self, page: int = None, page_size: int = None) -> Page[Project]:
        """
        List all projects.
        Args:
            page: the page number.
            page_size: the page size of each page.

        Returns:
            the projects list.

        Examples:
            .. code-block:: python
                import tidbcloudy
                api = tidbcloudy.TiDBCloud(public_key="your_public_key", private_key="your_private_key")
                projects = api.list_projects()
                total = projects.total
                for project in projects.items:
                    print(project) # This is a Project object

        """
        query = {}
        if page is not None:
            query["page"] = page
        if page_size is not None:
            query["page_size"] = page_size
        resp = self._context.call_get(path="projects", params=query)
        return Page(
            [Project.from_object(self._context, item) for item in resp["items"]],
            page, page_size, resp["total"])

    def iter_projects(self, page_size: int = 10) -> Iterator[Project]:
        """
        Iterate all projects.
        Args:
            page_size: the page size of each page.

        Returns:
            the projects iterator.

        Examples:
            .. code-block:: python
                import tidbcloudy
                api = tidbcloudy.TiDBCloud(public_key="your_public_key", private_key="your_private_key")
                for project in api.iter_projects():
                    print(project) # This is a Project object

        """
        page = 1
        total = None
        while total is None or (page - 1) * page_size < total:
            projects = self.list_projects(page=page, page_size=page_size)
            total = projects.total
            for project in projects.items:
                yield project
            page += 1

    def list_provider_regions(self) -> List[CloudSpecification]:
        """
        List all provider regions.
        Returns:
            the provider regions list.

        Examples:
            .. code-block:: python
                import tidbcloudy
                api = tidbcloudy.TiDBCloud(public_key="your_public_key", private_key="your_private_key")
                regions = api.list_provider_regions()
                for spec in api.list_provider_regions():
                    print(spec) # This is a CloudSpecification object

        """
        resp = self._context.call_get(path="clusters/provider/regions")
        return [CloudSpecification.from_object(obj=item) for item in resp["items"]]



from typing import TypeVar, Generic, List

T = TypeVar('T')


class Page(Generic[T]):
    def __init__(self, items: List[T], page: int, page_size: int, total: int):
        self._items = items
        self._page = page
        self._page_size = page_size
        self._total = total

    @property
    def items(self):
        return self._items

    @property
    def page(self):
        return self._page

    @property
    def page_size(self):
        return self._page_size

    @property
    def total(self):
        return self._total



from enum import Enum
from typing import Callable

from tidbcloudy.api.context import Context


class TiDBCloudyContextualBase:
    __slots__ = ["_context"]

    def __init__(self, context: Context, **kwargs):
        self._context = context
        super().__init__(**kwargs)

    @property
    def context(self):
        return self._context


class TiDBCloudyBase:
    _keys = None

    def __init_subclass__(cls):
        cls._keys = {}
        for key in cls.__slots__:
            key = key.removeprefix("_")
            cls._keys[key] = getattr(cls, key)

    def __init__(self, **kwargs):
        for key in self._keys:
            setattr(self, key, kwargs.pop(key, None))
        super().__init__(**kwargs)

    @classmethod
    def from_object(cls, context: Context = None, obj: dict = None):
        # Add context argument, and raise Exception when context is None
        #  and base class has TiDBCloudyContextualBase
        if issubclass(cls, TiDBCloudyContextualBase):
            if context is None:
                raise TypeError("context is None")
            inst = cls(context=context)
        else:
            inst = cls()
        inst.assign_object(obj)
        return inst

    def assign_object(self, obj: dict):
        # In from_object, forward context argument when self is subclass of TiDBCloudyContextualBase
        #  and forward a None context argument when self is not subclass of TiDBCloudyContextualBase
        context = self._context if isinstance(self, TiDBCloudyContextualBase) else None
        for key, descriptor in self._keys.items():
            value = obj.get(key, None)
            if value is None:
                continue
            if isinstance(descriptor, TiDBCloudyListField):
                setattr(self, key, [descriptor.item_type.from_object(context, item) for item in value])
            elif issubclass(descriptor.value_type, TiDBCloudyBase):
                setattr(self, key, descriptor.value_type.from_object(context, value))
            else:
                assert isinstance(descriptor, TiDBCloudyField)
                if descriptor.convert_from is not None:
                    value = descriptor.convert_from(value)
                setattr(self, key, value)

    def to_object(self) -> dict:
        obj = {}
        for key, descriptor in self._keys.items():
            value = getattr(self, key)
            if isinstance(descriptor, TiDBCloudyListField):
                if value is None:
                    value = []
                else:
                    value = [item.to_object() for item in value]
            else:
                assert isinstance(descriptor, TiDBCloudyField)
                if value is None and descriptor.none_is_empty:
                    continue
                if issubclass(descriptor.value_type, TiDBCloudyBase):
                    value = value.to_object()
                else:
                    if descriptor.convert_to is not None:
                        value = descriptor.convert_to(value)
            obj[key] = value
        return obj


class TiDBCloudyField:
    def __init__(self, value_type, convert_from: Callable = None, convert_to: Callable = None, none_is_empty=True):
        self._public = None
        self._private = None
        self.value_type = value_type
        self.convert_from = convert_from
        self.convert_to = convert_to
        self.none_is_empty = none_is_empty

        if issubclass(self.value_type, Enum):
            if self.convert_from is None:
                self.convert_from = lambda x: self.value_type(x.upper())
            if self.convert_to is None:
                self.convert_to = lambda x: x.value

    def __set_name__(self, owner, name):
        self._public = name
        self._private = "_" + name

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return getattr(instance, self._private, None)

    def __set__(self, instance, value):
        setattr(instance, self._private, value)


class TiDBCloudyListField:
    def __init__(self, item_type):
        self._public = None
        self._private = None
        self.item_type = item_type

    def __set_name__(self, owner, name):
        self._public = name
        self._private = "_" + name

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return getattr(instance, self._private)

    def __set__(self, instance, value):
        setattr(instance, self._private, value)



from enum import Enum
from typing import Union, List

from tidbcloudy.api._base import TiDBCloudyBase, TiDBCloudyField, TiDBCloudyListField
from tidbcloudy.util.ip import get_current_ip_address


class ClusterType(Enum):
    DEVELOPER = "DEVELOPER"
    DEDICATED = "DEDICATED"


class CloudProvider(Enum):
    AWS = "AWS"
    GCP = "GCP"


class ClusterStatus(Enum):
    INITIALIZING = "INITIALIZING"
    AVAILABLE = "AVAILABLE"
    CREATING = "CREATING"
    MODIFYING = "MODIFYING"
    PAUSED = "PAUSED"
    RESUMING = "RESUMING"
    UNAVAILABLE = "UNAVAILABLE"
    IMPORTING = "IMPORTING"
    CLEARED = "CLEARED"


class NodeStatus(Enum):
    NODE_STATUS_AVAILABLE = "NODE_STATUS_AVAILABLE"
    NODE_STATUS_UNAVAILABLE = "NODE_STATUS_UNAVAILABLE"
    NODE_STATUS_CREATING = "NODE_STATUS_CREATING"
    NODE_STATUS_DELETING = "NODE_STATUS_DELETING"


class BackupType(Enum):
    MANUAL = "MANUAL"
    AUTO = "AUTO"


class BackupStatus(Enum):
    PENDING = "PENDING"
    RUNNING = "RUNNING"
    FAILED = "FAILED"
    SUCCESS = "SUCCESS"


class RestoreStatus(Enum):
    PENDING = "PENDING"
    RUNNING = "RUNNING"
    FAILED = "FAILED"
    SUCCESS = "SUCCESS"


# noinspection PyShadowingBuiltins
class NodeQuantityRange(TiDBCloudyBase):
    __slots__ = ["_min", "_step"]
    min: int = TiDBCloudyField(int)
    step: int = TiDBCloudyField(int)


# noinspection PyShadowingBuiltins
class StorageSizeGibRange(TiDBCloudyBase):
    __slots__ = ["_min", "_max"]
    min: int = TiDBCloudyField(int)
    max: int = TiDBCloudyField(int)


class TiDBProfile(TiDBCloudyBase):
    __slots__ = ["_node_size", "_node_quantity_range"]
    node_size: str = TiDBCloudyField(str)
    node_quantity_range: NodeQuantityRange = TiDBCloudyField(NodeQuantityRange)

    def __repr__(self):
        return "<TiDB {}>".format(self.node_size)


class TiKVProfile(TiDBCloudyBase):
    __slots__ = ["_node_size", "_node_quantity_range", "_storage_size_gib_range"]
    node_size: str = TiDBCloudyField(str)
    node_quantity_range: NodeQuantityRange = TiDBCloudyField(NodeQuantityRange)
    storage_size_gib_range: StorageSizeGibRange = TiDBCloudyField(StorageSizeGibRange)

    def __repr__(self):
        return "<TiKV {}>".format(self.node_size)


class TiFlashProfile(TiDBCloudyBase):
    __slots__ = ["_node_size", "_node_quantity_range", "_storage_size_gib_range"]
    node_size: str = TiDBCloudyField(str)
    node_quantity_range: NodeQuantityRange = TiDBCloudyField(NodeQuantityRange)
    storage_size_gib_range: StorageSizeGibRange = TiDBCloudyField(StorageSizeGibRange)

    def __repr__(self):
        return "<TiFlash {}>".format(self.node_size)


class ConnectionString(TiDBCloudyBase):
    __slots__ = ["_host", "_port"]
    host: str = TiDBCloudyField(str)
    port: int = TiDBCloudyField(int)


class ConnectionStrings(TiDBCloudyBase):
    __slots__ = ["_default_user", "_standard", "_vpc_peering"]
    default_user: str = TiDBCloudyField(str)
    standard: ConnectionString = TiDBCloudyField(ConnectionString)
    vpc_peering: ConnectionString = TiDBCloudyField(ConnectionString)


class TiDBNodeMap(TiDBCloudyBase):
    __slots__ = ["_node_name", "_availability_zone", "_node_size", "_vcpu_num", "_ram_bytes", "_status"]
    node_name: str = TiDBCloudyField(str)
    availability_zone: str = TiDBCloudyField(str)
    node_size: str = TiDBCloudyField(str)
    vcpu_num: int = TiDBCloudyField(int)
    ram_bytes: int = TiDBCloudyField(int)
    status: NodeStatus = TiDBCloudyField(NodeStatus)


class TiKVNodeMap(TiDBCloudyBase):
    __slots__ = [
        "_node_name", "_availability_zone", "_node_size", "_vcpu_num", "_ram_bytes", "_storage_size_gib", "_status"]

    node_name: str = TiDBCloudyField(str)
    availability_zone: str = TiDBCloudyField(str)
    node_size: str = TiDBCloudyField(str)
    vcpu_num: int = TiDBCloudyField(int)
    ram_bytes: int = TiDBCloudyField(int)
    storage_size_gib: int = TiDBCloudyField(int)
    status: NodeStatus = TiDBCloudyField(NodeStatus)


class TiFlashNodeMap(TiDBCloudyBase):
    __slots__ = ["_node_name", "_availability_zone", "_node_size", "_vcpu_num", "_ram_bytes", "_storage_size_gib",
                 "_status", ]

    node_name: str = TiDBCloudyField(str)
    availability_zone: str = TiDBCloudyField(str)
    node_size: str = TiDBCloudyField(str)
    vcpu_num: int = TiDBCloudyField(int)
    ram_bytes: int = TiDBCloudyField(int)
    storage_size_gib: int = TiDBCloudyField(int)
    status: NodeStatus = TiDBCloudyField(NodeStatus)


class NodeMapSpec(TiDBCloudyBase):
    __slots__ = ["_tidb", "_tikv", "_tiflash"]
    tidb: List[TiDBNodeMap] = TiDBCloudyListField(TiDBNodeMap)
    tikv: List[TiKVNodeMap] = TiDBCloudyListField(TiKVNodeMap)
    tiflash: List[TiFlashNodeMap] = TiDBCloudyListField(TiFlashNodeMap)


class IPAccessList(TiDBCloudyBase):
    __slots__ = ["_cidr", "_description"]
    cidr: str = TiDBCloudyField(str)
    description: str = TiDBCloudyField(str)


class TiDBComponent(TiDBCloudyBase):
    __slots__ = ["_node_size", "_node_quantity"]
    node_size: str = TiDBCloudyField(str)
    node_quantity: int = TiDBCloudyField(int)


class TiKVComponent(TiDBCloudyBase):
    __slots__ = ["_node_size", "_node_quantity", "_storage_size_gib"]
    node_size: str = TiDBCloudyField(str)
    node_quantity: int = TiDBCloudyField(int)
    storage_size_gib: int = TiDBCloudyField(int)


class TiFlashComponent(TiDBCloudyBase):
    __slots__ = ["_node_size", "_node_quantity", "_storage_size_gib"]
    node_size: str = TiDBCloudyField(str)
    node_quantity: int = TiDBCloudyField(int)
    storage_size_gib: int = TiDBCloudyField(int)


class ClusterComponents(TiDBCloudyBase):
    __slots__ = ["_tidb", "_tikv", "_tiflash"]
    tidb: TiDBComponent = TiDBCloudyField(TiDBComponent)
    tikv: TiKVComponent = TiDBCloudyField(TiKVComponent)
    tiflash: TiFlashComponent = TiDBCloudyField(TiFlashComponent)


class ClusterConfig(TiDBCloudyBase):
    __slots__ = ["_root_password", "_port", "_components", "_ip_access_list"]
    root_password: str = TiDBCloudyField(str)
    port: int = TiDBCloudyField(int)
    components: ClusterComponents = TiDBCloudyField(ClusterComponents)
    ip_access_list: List[IPAccessList] = TiDBCloudyListField(IPAccessList)

    def __repr__(self):
        return "<ClusterConfig port={} components={}>".format(self.port, self.components)


class CreateClusterConfig:
    def __init__(self):
        self._name = None
        self._cluster_type = None
        self._cloud_provider = None
        self._region = None
        self._root_password = None
        self._port = None
        self._components = None
        self._tidb = None
        self._tikv = None
        self._tiflash = None
        self._ip_access_list = []

    def set_name(self, name: str):
        self._name = name
        return self

    def set_cluster_type(self, cluster_type: Union[ClusterType, str]):
        if isinstance(cluster_type, ClusterType):
            self._cluster_type = cluster_type.value
        else:
            self._cluster_type = cluster_type.upper()
        return self

    def set_cloud_provider(self, cloud_provider: Union[CloudProvider, str]):
        if isinstance(cloud_provider, CloudProvider):
            self._cloud_provider = cloud_provider.value
        else:
            self._cloud_provider = cloud_provider.upper()
        return self

    def set_region(self, region: str):
        self._region = region
        return self

    def set_root_password(self, root_password: str):
        self._root_password = root_password
        return self

    def set_port(self, port: int):
        self._port = port
        return self

    def set_component(self, component: str, node_size: str, node_quantity: int, storage_size_gib: int = None):
        component = component.lower()
        if component == "tidb":
            self._tidb = TiDBComponent(node_size=node_size, node_quantity=node_quantity)
        elif component == "tikv":
            self._tikv = TiKVComponent(node_size=node_size, node_quantity=node_quantity,
                                       storage_size_gib=storage_size_gib)
        elif component == "tiflash":
            self._tiflash = TiFlashComponent(node_size=node_size, node_quantity=node_quantity,
                                             storage_size_gib=storage_size_gib)
        else:
            raise Exception("Invalid component, only tidb, tikv, tiflash are supported")
        return self

    def add_ip_access(self, cidr: str, description: str = ""):
        self._ip_access_list.append(IPAccessList(cidr=cidr, description=description))
        return self

    def add_current_ip_access(self, description: str = "Current IP automatically generated by tidbcloudy"):
        self.add_ip_access(cidr=get_current_ip_address() + "/32", description=description)
        return self

    def to_object(self) -> dict:
        return {
            "name": self._name if self._name is not None else "",
            "cluster_type": self._cluster_type,
            "cloud_provider": self._cloud_provider,
            "region": self._region,
            "config": {
                "root_password": self._root_password,
                "port": self._port,
                "components": {
                    "tidb": {
                        "node_size": self._tidb.node_size,
                        "node_quantity": self._tidb.node_quantity
                    } if self._tidb is not None else None,
                    "tikv": {
                        "node_size": self._tikv.node_size,
                        "node_quantity": self._tikv.node_quantity,
                        "storage_size_gib": self._tikv.storage_size_gib
                    } if self._tikv is not None else None,
                    "tiflash": {
                        "node_size": self._tiflash.node_size,
                        "node_quantity": self._tiflash.node_quantity,
                        "storage_size_gib": self._tiflash.storage_size_gib
                    } if self._tiflash is not None else None
                },
                "ip_access_list": [
                    item.to_object() for item in self._ip_access_list] if self._ip_access_list is not None else None
            }
        }


class UpdateClusterConfig:
    def __init__(self):
        self._tidb = None
        self._tikv = None
        self._tiflash = None

    def update_component(self, component: str, node_quantity: int = None, node_size: str = None,
                         storage_size_gib: int = None):
        if component == "tidb" and node_quantity is not None:
            self._tidb = TiDBComponent(node_size=node_size, node_quantity=node_quantity)
        elif component == "tikv" and node_quantity is not None:
            self._tikv = TiKVComponent(node_size=node_size, node_quantity=node_quantity,
                                       storage_size_gib=storage_size_gib)
        elif component == "tiflash" and node_quantity is not None:
            self._tiflash = TiFlashComponent(node_size=node_size, node_quantity=node_quantity,
                                             storage_size_gib=storage_size_gib)
        else:
            raise Exception("Invalid component, only tidb, tikv, tiflash are supported")
        return self

    def to_object(self) -> dict:
        components = {}
        if self._tidb is not None:
            components["tidb"] = {}
            if self._tidb.node_size is not None:
                components["tidb"]["node_size"] = self._tidb.node_size
            if self._tidb.node_quantity is not None:
                components["tidb"]["node_quantity"] = self._tidb.node_quantity
        if self._tikv is not None:
            components["tikv"] = {}
            if self._tikv.node_size is not None:
                components["tikv"]["node_size"] = self._tikv.node_size
            if self._tikv.node_quantity is not None:
                components["tikv"]["node_quantity"] = self._tikv.node_quantity
            if self._tikv.storage_size_gib is not None:
                components["tikv"]["storage_size_gib"] = self._tikv.storage_size_gib
        if self._tiflash is not None:
            components["tiflash"] = {}
            if self._tiflash.node_size is not None:
                components["tiflash"]["node_size"] = self._tiflash.node_size
            if self._tiflash.node_quantity is not None:
                components["tiflash"]["node_quantity"] = self._tiflash.node_quantity
            if self._tiflash.storage_size_gib is not None:
                components["tiflash"]["storage_size_gib"] = self._tiflash.storage_size_gib

        return {
            "config": {
                "components": components
            }
        }


class ClusterInfo(TiDBCloudyBase):
    __slots__ = ["_tidb_version", "_cluster_status", "_node_map", "_connection_strings"]
    tidb_version: str = TiDBCloudyField(str)
    cluster_status: ClusterStatus = TiDBCloudyField(ClusterStatus)
    node_map: NodeMapSpec = TiDBCloudyField(NodeMapSpec)
    connection_strings: ConnectionStrings = TiDBCloudyField(ConnectionStrings)

    def __repr__(self):
        return "<tidb={} status= {} default user={}>".format(
            self.tidb_version,
            self.cluster_status.value if self.cluster_status is not None else None,
            self.connection_strings.default_user)


# noinspection PyShadowingBuiltins
class ClusterInfoOfRestore(TiDBCloudyBase):
    __slots__ = ["_id", "_name", "_status", ]
    id: str = TiDBCloudyField(str)
    name: str = TiDBCloudyField(str)
    status: ClusterStatus = TiDBCloudyField(ClusterStatus)

    def __repr__(self):
        return "<id={}, name={}, status={}>".format(self.id, self.name,
                                                    self.status.value if self.status is not None else None)


class CloudSpecification(TiDBCloudyBase):
    __slots__ = ["_cluster_type", "_cloud_provider", "_region", "_tidb", "_tikv", "_tiflash"]
    cluster_type: ClusterType = TiDBCloudyField(ClusterType)
    cloud_provider: CloudProvider = TiDBCloudyField(CloudProvider)
    region: str = TiDBCloudyField(str)
    tidb: List[TiDBProfile] = TiDBCloudyListField(TiDBProfile)
    tikv: List[TiKVProfile] = TiDBCloudyListField(TiKVProfile)
    tiflash: List[TiFlashProfile] = TiDBCloudyListField(TiFlashProfile)

    def __repr__(self):
        return "<Specification cluster_type={} cloud_provider={} region={}>".format(
            self.cluster_type.value if self.cluster_type is not None else None,
            self.cloud_provider.value if self.cloud_provider is not None else None,
            self.region)



from tidbcloudy.api._base import TiDBCloudyBase, TiDBCloudyContextualBase, TiDBCloudyField
from .specification import BackupType, BackupStatus
from tidbcloudy.util.log import log


# noinspection PyShadowingBuiltins
class Backup(TiDBCloudyBase, TiDBCloudyContextualBase):
    __slots__ = ["_id", "_cluster_id", "_project_id", "_name", "_description", "_type", "_create_timestamp", "_size",
                 "_status"]

    id: str = TiDBCloudyField(str)
    cluster_id: str = TiDBCloudyField(str)
    project_id: str = TiDBCloudyField(str)
    name: str = TiDBCloudyField(str)
    description: str = TiDBCloudyField(str)
    type: BackupType = TiDBCloudyField(BackupType)
    create_timestamp: str = TiDBCloudyField(str)
    size: str = TiDBCloudyField(str)
    status: BackupStatus = TiDBCloudyField(BackupStatus)

    """
    Args:
        id: the id of the backup
        cluster_id: the id of the cluster
        name: the name of the backup
        description: the description of the backup
        type: the type of the backup
        create_timestamp: the created timestamp of the backup
        size: the size of the backup
        status: the status of the backup
        _from: internal use only
    """

    def delete(self):
        path = "projects/{}/clusters/{}/backups/{}".format(self.project_id, self.cluster_id, self.id)
        self.context.call_delete(path=path)
        log("backup task id={} has been deleted".format(self.id))

    def __repr__(self):
        return "<backup id={} name={} create_at= {}>".format(self.id, self.name, self.create_timestamp)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tidbcloudy', 'tidbcloudy.api', 'tidbcloudy.exception', 'tidbcloudy.util']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.3,<9.0.0',
 'deprecation>=2.1.0,<3.0.0',
 'mysqlclient>=2.1.1,<3.0.0',
 'requests>=2.27.1,<3.0.0',
 'tabulate>=0.9.0,<0.10.0',
 'toml>=0.10.2,<0.11.0']

entry_points = \
{'console_scripts': ['tidbcloudy = tidbcloudy.__main__:cli']}

setup_kwargs = {
    'name': 'tidbcloudy',
    'version': '1.1.0a1',
    'description': 'Python SDK for TiDB Cloud',
    'long_description': '# Python SDK for TiDB Cloud\n\n`tidbcloudy` is an unofficial Python SDK for [TiDB Cloud](https://tidbcloud.com).\n\n## Introduction\n\nFor more information about TiDB Cloud API, see [TiDB Cloud API Documentation](https://docs.pingcap.com/tidbcloud/api/v1beta).\n\n> TiDB Cloud is a fully-managed Database-as-a-Service (DBaaS) that brings everything great about TiDB to your cloud.\n\nIf you do not have a TiDB Cloud account yet, you can sign up [here](https://tidbcloud.com). For more details about TiDB Cloud, refer to [TiDB Cloud Documentation](https://docs.pingcap.com/tidbcloud/).\n\nYou can use this SDK to access [TiDB Cloud](https://tidbcloud.com) and manage your projects, clusters, backups and restores:\n\n- manage your TiDB Cloud **projects** (only _list_ is supported now)\n- list all available cloud providers (AWS and GCP), regions and specifications before creating or modifying a cluster\n- manage your Serverless Tier or Dedicated Tier **clusters** (_create_, _modify_, _pause_, _resume_, _get_, _list_, _delete_)\n- manage your **backups** of a cluster (_create_, _get_, _list_, _delete_)\n- manage your **restores** of a project (_create_, _get_, _list_)\n\n### Compatibility with TiDB Cloud API\n\n`tidbcloudy` is compatible with [TiDB Cloud API](https://docs.pingcap.com/tidbcloud/api/v1beta). The following table lists the supported API versions:\n\n| tidbcloudy                                                         | TiDB Cloud API                                                                                          |\n|--------------------------------------------------------------------|---------------------------------------------------------------------------------------------------------|\n| [1.0.1](https://github.com/Oreoxmt/tidbcloudy/releases/tag/v1.0.1) | v1beta [Release 20220906](https://docs.pingcap.com/tidbcloud/api/v1beta#section/API-Changelog/20220906), [Release 20220920](https://docs.pingcap.com/tidbcloud/api/v1beta#section/API-Changelog/20220920), [Release 20221028](https://docs.pingcap.com/tidbcloud/api/v1beta#section/API-Changelog/20221028)|\n| [1.0.0](https://github.com/Oreoxmt/tidbcloudy/releases/tag/v1.0.0) | v1beta [Release 20220823](https://docs.pingcap.com/tidbcloud/api/v1beta#section/API-Changelog/20220823) |\n| [0.2.1](https://github.com/Oreoxmt/tidbcloudy/releases/tag/v0.2.1) | v1beta [Release 20220809](https://docs.pingcap.com/tidbcloud/api/v1beta#section/API-Changelog/20220809) |                                                                          |\n\n### Enhancements comparing to original [TiDB Cloud API](https://docs.pingcap.com/tidbcloud/api/v1beta)\n\n- **Iterate over** resources instead of manual pagination\n- **Connect to a TiDB cluster** using the [MySQL client](https://github.com/PyMySQL/mysqlclient)\n- **Get** a Project using a Project **ID**\n- Configure your cluster with **method chaining**\n- Add your **current IP address automatically**\n- **Wait for the cluster to be ready** when creating/modifying a cluster\n- **Case-insensitive** when setting cluster type, cloud provider, and component name\n\n## Installation\n\n```bash\npip3 install tidbcloudy\n```\n\n⚠️ Make sure that you have installed **mysql client** in your environment. For more information, see [PyMySQL/mysqlclient](https://github.com/PyMySQL/mysqlclient#install).\n\n## Usage\n\n### Prerequisites\n\n- Create a TiDB Cloud account.\n- Create a TiDB Cloud API key. To manage your API keys, see [TiDB Cloud API Documentation](https://docs.pingcap.com/tidbcloud/api/v1beta#section/Authentication/API-Key-Management).\n- Install the latest version of `tidbcloudy`.\n\nTo get full code examples, see the [`examples`](https://github.com/Oreoxmt/tidbcloudy/tree/main/examples) folder.\n\n> Note:\n>\n> It is recommended to set environment variables for your API public and private key. For example, in bash, you can:\n>\n> ```bash\n> export PUBLIC_KEY=your_api_public_key\n> export PRIVATE_KEY=your_api_private_key\n> ```\n\n### List all resources in your organization\n\nTo get the full code example of listing all projects, clusters, backup tasks, and restore tasks in your organization, see [`0_list_resources.py`](https://github.com/Oreoxmt/tidbcloudy/tree/main/examples/0_list_resources.py).\n\n```python\nimport os\n\nimport tidbcloudy\nfrom tidbcloudy.api.specification import ClusterType\n\npublic_key = os.environ.get("PUBLIC_KEY")\nprivate_key = os.environ.get("PRIVATE_KEY")\n\napi = tidbcloudy.TiDBCloud(public_key=public_key, private_key=private_key)\n\nfor project in api.iter_projects():\n    print(project)\n    for cluster in project.iter_clusters():\n        print(cluster)\n        if cluster.cluster_type == ClusterType.DEDICATED:\n            for backup in cluster.iter_backups():\n                print(backup)\n    for restore in project.iter_restores():\n        print(restore)\n```\n\n### Create a cluster\n\nBefore creating a cluster, you should list all available provider regions and cluster configuration specifications. For more details, run the [`1_list_provider_regions.py`](https://github.com/Oreoxmt/tidbcloudy/tree/main/examples/1_list_provider_regions.py).\n\n```python\nimport os\n\nimport tidbcloudy\n\npublic_key = os.environ.get("PUBLIC_KEY")\nprivate_key = os.environ.get("PRIVATE_KEY")\n\napi = tidbcloudy.TiDBCloud(public_key=public_key, private_key=private_key)\n\nfor spec in api.list_provider_regions():\n    print(f"- type: {spec.cluster_type.value}")\n    print(f"  provider: {spec.cloud_provider.value}")\n    print(f"  region: {spec.region}")\n    print(f"  components:")\n    for tidb in spec.tidb:\n        print(f"  - tidb: {tidb.node_size}; "\n              f"min={tidb.node_quantity_range.min} step={tidb.node_quantity_range.step}")\n    for tikv in spec.tikv:\n        print(f"  - tikv: {tikv.node_size}; "\n              f"min={tikv.node_quantity_range.min} "\n              f"step={tikv.node_quantity_range.step}; "\n              f"{tikv.storage_size_gib_range.min}..{tikv.storage_size_gib_range.max} GiB")\n    for tiflash in spec.tiflash:\n        print(\n            f"  - tiflash: {tiflash.node_size}; "\n            f"min={tiflash.node_quantity_range.min} step={tiflash.node_quantity_range.step}; "\n            f"{tiflash.storage_size_gib_range.min}..{tiflash.storage_size_gib_range.max} GiB")\n```\n\n> Note:\n>\n> Creating a cluster might cost money. For more details, see [TiDB Cloud pricing details](https://www.pingcap.com/tidb-cloud-pricing-details/).\n\nTo create a Serverless Tier cluster, run the [`2_1_create_serverless_cluster.py`](https://github.com/Oreoxmt/tidbcloudy/tree/main/examples/2_1_create_serverless_cluster.py).\n\nTo create a Dedicated Tier cluster, run the [`2_2_create_dedicated_cluster.py`](https://github.com/Oreoxmt/tidbcloudy/tree/main/examples/2_2_create_dedicated_cluster.py).\n\nThe following takes creating a Serverless Tier cluster as an example:\n\n```python\nimport os\nimport tidbcloudy\nfrom tidbcloudy.api.specification import CreateClusterConfig\n\npublic_key = os.environ.get("PUBLIC_KEY")\nprivate_key = os.environ.get("PRIVATE_KEY")\ndebug_mode = os.environ.get("TIDBCLOUDY_LOG")\nproject_id = "1234567890123456789"\n\napi = tidbcloudy.TiDBCloud(public_key=public_key, private_key=private_key)\nproject = api.get_project(project_id, update_from_server=True)\n\nconfig = CreateClusterConfig()\nconfig \\\n    .set_name("serverless-0") \\\n    .set_cluster_type("DEVELOPER") \\\n    .set_cloud_provider("AWS") \\\n    .set_region("us-west-2") \\\n    .set_root_password("your_root_password") \\\n    .add_ip_access(cidr="0.0.0.0/0") \\\n    .add_current_ip_access()\ncluster = project.create_cluster(config)\nprint(cluster)\n\ncluster.wait_for_ready()\n```\n\n### Connect to TiDB\n\nTo connect to your TiDB cluster, run the [`3_connect_mysql.py`](https://github.com/Oreoxmt/tidbcloudy/tree/main/examples/3_connect_mysql.py).\n\n```python\nimport os\n\nimport tidbcloudy\nfrom tidbcloudy.api.specification import ClusterStatus\n\npublic_key = os.environ.get("PUBLIC_KEY")\nprivate_key = os.environ.get("PRIVATE_KEY")\nproject_id = "1234567890123456789"\ncluster_id = "1234567890123456789"\n\nprint("Connecting to TiDB Cloud...")\napi = tidbcloudy.TiDBCloud(public_key=public_key, private_key=private_key)\nproject = api.get_project(project_id, update_from_server=True)\ncluster = project.get_cluster(cluster_id)\nprint(cluster)\n\nif cluster.status.cluster_status == ClusterStatus.AVAILABLE:\n    connection_strings = cluster.status.connection_strings\n    connection = cluster.connect(type="standard", database="test", password="your_root_password")\n    print(connection)\n    with connection:\n        with connection.cursor() as cursor:\n            cursor.execute("SELECT DATABASE();")\n            m = cursor.fetchone()\n            print(m[0])\n```\n\n### Modify a cluster\n\n> Note:\n>\n> Modify a cluster might cost money. For more details, see [TiDB Cloud pricing details](https://www.pingcap.com/tidb-cloud-pricing-details/).\n\nTo modify a cluster, run the [`4_scale_a_cluster.py`](https://github.com/Oreoxmt/tidbcloudy/tree/main/examples/4_scale_a_cluster.py).\n\n```python\nimport os\n\nimport tidbcloudy\nfrom tidbcloudy.api.specification import UpdateClusterConfig\n\npublic_key = os.environ.get("PUBLIC_KEY")\nprivate_key = os.environ.get("PRIVATE_KEY")\nproject_id = "1234567890123456789"\ncluster_id = "1234567890123456789"\n\napi = tidbcloudy.TiDBCloud(public_key=public_key, private_key=private_key)\nproject = api.get_project(project_id, update_from_server=True)\ncluster = project.get_cluster(cluster_id)\n\nprint("The original config is: {}".format(cluster.config.components.to_object()))\n\nnew_config = UpdateClusterConfig()\nnew_config.update_component("tiflash", node_quantity=1, node_size="8C64G", storage_size_gib=500)\ncluster.update(new_config)\ncluster.wait_for_ready()\n\nprint("The new config is: {}".format(cluster.config.components.to_object()))\n```\n\n### Backup and restore\n\n> Note:\n>\n> Backup or restore a cluster might cost money. For more details, see [TiDB Cloud pricing details](https://www.pingcap.com/tidb-cloud-pricing-details/).\n\nTo create a backup and restore, run the [`5_backup_restore.py`](https://github.com/Oreoxmt/tidbcloudy/tree/main/examples/5_backup_restore.py)\n\n```python\nimport os\n\nimport tidbcloudy\nfrom tidbcloudy.api.specification import CreateClusterConfig\n\npublic_key = os.environ.get("PUBLIC_KEY")\nprivate_key = os.environ.get("PRIVATE_KEY")\nproject_id = "1234567890123456789"\ncluster_id = "1234567890123456789"\nbackup_id = "1234567"\n\napi = tidbcloudy.TiDBCloud(public_key=public_key, private_key=private_key)\nproject = api.get_project(project_id, update_from_server=True)\ncluster = project.get_cluster(cluster_id)\nprint("Create a manual backup task")\nbackup = cluster.create_backup(name="backup-1", description="automatically generated by tidbcloudy")\nprint(backup)\n\nconfig = CreateClusterConfig()\nconfig \\\n    .set_cluster_type("DEDICATED") \\\n    .set_cloud_provider("AWS") \\\n    .set_region("us-west-2") \\\n    .set_port(4399) \\\n    .set_root_password("your_root_password") \\\n    .set_component("tidb", "8C16G", 1) \\\n    .set_component("tikv", "8C32G", 3, 500) \\\n    .set_component("tiflash", "8C64G", 1, 500) \\\n    .add_current_ip_access()\nprint("Create a restore task from backup_id={}".format(backup_id))\nrestore = project.create_restore(backup_id=backup_id, name="restore-1", cluster_config=config)\nrestore_task = project.get_restore(restore.id)\nprint(restore_task.to_object())\nfor restore in project.iter_restores():\n    print(restore)\n```\n\n### Pause or resume your cluster\n\nTo pause or resume your cluster, run the [`6_pause_cluster.py`](https://github.com/Oreoxmt/tidbcloudy/tree/main/examples/6_pause_cluster.py).\n\n```python\nimport os\n\nimport tidbcloudy\nfrom tidbcloudy.api.specification import ClusterStatus\n\npublic_key = os.environ.get("PUBLIC_KEY")\nprivate_key = os.environ.get("PRIVATE_KEY")\nproject_id = "1234567890123456789"\ncluster_id = "1234567890123456789"\n\napi = tidbcloudy.TiDBCloud(public_key=public_key, private_key=private_key)\nproject = api.get_project(project_id, update_from_server=True)\ncluster = project.get_cluster(cluster_id)\n\nif cluster.status.cluster_status == ClusterStatus.AVAILABLE:\n    print("Pause the cluster id={}".format(cluster_id))\n    cluster.pause()\nif cluster.status.cluster_status == ClusterStatus.PAUSED:\n    print("Resume the cluster id={}".format(cluster_id))\n    cluster.resume()\nif cluster.status.cluster_status == ClusterStatus.RESUMING:\n    print("Wait for the RESUMING cluster id={} to be available".format(cluster_id))\n    cluster.wait_for_available()\n```\n\n### Delete all resources\n\n> Warning:\n>\n> This is a destructive operation. It will delete all resources in the project. **DO NOT** run this script in a production environment.\n\nTo delete all clusters and backup tasks in your project, run the [`7_delete_resources.py`](https://github.com/Oreoxmt/tidbcloudy/tree/main/examples/7_delete_resources.py).\n\n```python\nimport os\n\nimport tidbcloudy\nfrom tidbcloudy.api.specification import ClusterType\n\npublic_key = os.environ.get("PUBLIC_KEY")\nprivate_key = os.environ.get("PRIVATE_KEY")\nproject_id = "1234567890123456789"\n\napi = tidbcloudy.TiDBCloud(public_key=public_key, private_key=private_key)\nproject = api.get_project(project_id, update_from_server=True)\nfor cluster in project.iter_clusters():\n    print(cluster)\n    if cluster.cluster_type == ClusterType.DEDICATED:\n        for backup in cluster.iter_backups():\n            print(backup)\n            backup.delete()\n    cluster.delete()\n```\n\n## Related Projects\n\n- Go SDK: [go-tidbcloud-sdk-v1](https://github.com/c4pt0r/go-tidbcloud-sdk-v1) by [@c4pt0r](https://github.com/c4pt0r)\n- TiDB Cloud CLI: [OhMyTiUP/tidb-cloud](https://github.com/luyomo/OhMyTiUP/tree/main/pkg/tidbcloudapi) by [@luyomo](https://github.com/luyomo)\n- Official code samples in Go and Python: [tidbcloud-api-samples](https://github.com/tidbcloud/tidbcloud-api-samples)\n',
    'author': 'Aolin',
    'author_email': 'aolinz@outlook.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
